import os

from moviepy.editor import VideoFileClip
from moviepy.editor import concatenate_videoclips

from yt_concate_py.pipeline.steps.step import Step
from yt_concate_py.settings import VIDEOS_DIR
from yt_concate_py.settings import OUTPUTS_DIR
from yt_concate_py.logger import logger


class EditVideo(Step):
    def process(self, data, inputs, utils):
        logger.info('EDITING VIDEO...')
        clips = []
        i = 0
        for found in data:
            logger.debug(f'{found.yt.video_link}')
            logger.debug(f'{found.time}')
            start, end = self.parse_caption_time(found.time)
            logger.debug(f'{start, end}')
            filepath = os.path.join(VIDEOS_DIR, found.yt.id + '.mp4')
            if not os.path.exists(filepath):
                logger.info(f'video file not found: {filepath}')
                continue
            try:
                video = VideoFileClip(filepath)
                if end > video.duration:
                    logger.info(f'end time exceeds video duration for {filepath}, adjusting {end} to {video.duration}')
                    end = video.duration
                video = VideoFileClip(filepath).subclip(start, end)
                clips.append(video)
                i += 1
                logger.info(f'clip has been added {filepath}, {start, end}, {i}')
                if len(clips) >= int(inputs['limit']):
                    break
            except Exception as e:
                logger.info(f'error for editing video {filepath}, {start, end} and the error is {str(e)}')
        final_clip = concatenate_videoclips(clips, method="compose")
        temp_audio_filepath = os.path.join(OUTPUTS_DIR, inputs['channel_id'] + '_' + inputs['search_term'] + 'TEMP.mp4')
        output_filepath = os.path.join(OUTPUTS_DIR, inputs['channel_id'] + '_' + inputs['search_term'] + '.mp4')
        final_clip.write_videofile(
            output_filepath,
            remove_temp=False,
            temp_audiofile=temp_audio_filepath,
            audio_codec="aac",
            codec='libx264',
        )
        logger.info('Completed editing video')

    def parse_caption_time(self, caption_time):
        start, end = caption_time.split(' --> ')
        return self.parse_time_str(start), self.parse_time_str(end)

    @staticmethod
    def parse_time_str(time_str):
        h, m, s = time_str.split(':')
        s, ms = s.split(',')
        return float(h) * 3600 + float(m) * 60 + float(s) + float(ms) / 1000



import logging

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(levelname)s:%(asctime)s:%(message)s')
file_handler = logging.FileHandler('yt_concate_py.log')
file_handler.setLevel(logging.DEBUG)
file_handler.setFormatter(formatter)
logger.addHandler(file_handler)

stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)



class Found:
    def __init__(self, yt, caption, time):
        self.yt = yt
        self.caption = caption
        self.time = time

    def __str__(self):
        return '<Found(yt=' + str(self.yt) + ')>'

    def __repr__(self):
        content = ':'.join(
            [
                'yt=' + str(self.yt),
                'caption=' + str(self.caption),
                'time=' + str(self.time)
            ]
        )
        return '<Found(' + content + ')>'



import time
import os

from yt_concate_py.pipeline.steps.step import Step
from yt_concate_py.settings import CAPTIONS_DIR
from yt_concate_py.logger import logger


class ReadCaptions(Step):
    def process(self, data, inputs, utils):
        logger.info('READING CAPTIONS...')
        for yt in data:
            logger.debug(f'reading yt : {yt.video_link}')
            filepath = os.path.join(CAPTIONS_DIR, yt.id + '.txt')
            is_file_exist = os.path.exists(filepath)
            if is_file_exist is not True:
                logger.debug(f'{is_file_exist}! caption file not found : {filepath}')
                continue
            captions = {}
            logger.debug(f'{is_file_exist}! caption file exists : {filepath}')
            with open(filepath, 'r', encoding='utf-8') as f:
                is_time = False
                timestamp = None
                caption = None
                for line in f:
                    line = line.strip()
                    if "-->" in line:
                        timestamp = line
                        is_time = True
                        continue
                    if is_time:
                        caption = line
                        is_time = False
                        captions[caption] = timestamp
            yt.captions = captions

        return data



from yt_concate_py.pipeline.steps.step import Step
from yt_concate_py.logger import logger


class Postflight(Step):
    def process(self, data, inputs, utils):
        logger.info('IN POSTFLIGHT...')



import os

from yt_concate_py.settings import CAPTIONS_DIR
from yt_concate_py.settings import VIDEOS_DIR


class YT:
    def __init__(self, video_link):
        self.video_link = video_link
        self.id = self.get_video_id(self.video_link)
        self.caption_path = self.get_caption_path()
        self.caption_file_exist = self.caption_file_exist()
        self.video_path = self.get_video_path()
        self.video_filename = self.get_video_filename()
        self.video_file_exist = self.video_file_exist()
        self.captions = None


    @staticmethod
    def get_video_id(video_link):
        return video_link.split("?v=")[1]

    def get_caption_path(self):
        return os.path.join(CAPTIONS_DIR, self.id + '.txt')

    def caption_file_exist(self):
        path = self.caption_path
        return os.path.exists(path) and os.path.getsize(path) > 0

    def get_video_path(self):
        return os.path.join(VIDEOS_DIR, self.id + '.mp4')

    def get_video_filename(self):
        return self.id + '.mp4'

    def video_file_exist(self):
        path = self.video_path
        return os.path.exists(path) and os.path.getsize(path) > 0

    def __str__(self):
        return '<YT(' + self.video_link + ')>'

    def __repr__(self):
        content = ':'.join(
            [
                'id=' + str(self.id),
                'caption_path=' + str(self.caption_path),
                'video_path=' + str(self.video_path)
            ]
        )
        return '<found(' + content + ')>'



from yt_concate_py.pipeline.steps.step import Step
from yt_concate_py.logger import logger


class Preflight(Step):
    def process(self, data, inputs, utils):
        logger.info('IN PREFLIGHT...')
        channel_id = inputs['channel_id']
        search_term = inputs['search_term']
        logger.info(f'Prepare to concatenate video clips from YT channel {channel_id} with {search_term}...')
        utils.create_dirs()



import os
from dotenv import load_dotenv


load_dotenv()
API_KEY = os.getenv('API_KEY')

DOWNLOADS_DIR = 'downloads'
VIDEOS_DIR = os.path.join(DOWNLOADS_DIR, 'videos')
CAPTIONS_DIR = os.path.join(DOWNLOADS_DIR, 'captions')
OUTPUTS_DIR = 'outputs'



import time
import os
import concurrent.futures

from pytubefix import YouTube

from yt_concate_py.pipeline.steps.step import Step
from yt_concate_py.settings import CAPTIONS_DIR
from yt_concate_py.logger import logger


class DownloadCaptions(Step):
    def process(self, data, inputs, utils):
        logger.info('DOWNLOADING CAPTIONS...')
        start_time = time.time()
        yt_ = []
        for yt in data:
            if yt.caption_file_exist:
                logger.debug('found existing caption file')
                continue
            else:
                yt_.append(yt)

        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
            executor.map(self.download_captions, yt_)

        end_time = time.time()
        elapsed_time = end_time - start_time
        logger.info('Completed downloading captions')
        logger.info(f'Downloaded {len(yt_)} new caption files')
        logger.info(f'Total running time for downloading captions : {elapsed_time} seconds')
        return data

    def download_captions(self, yt):
        logger.debug(f'downloading caption : {yt.video_link}')
        youtube = YouTube(yt.video_link)
        subtitles = list(youtube.captions)
        try:
            first_subtitle = subtitles[0]
            first_subtitle_code = first_subtitle.code
            caption = youtube.captions[first_subtitle_code]
            caption_srt = caption.generate_srt_captions()

            filepath = os.path.join(CAPTIONS_DIR, yt.id + '.txt')
            try:
                logger.debug(f'writing caption file for {filepath}')
                with open(filepath, 'w', encoding='utf-8') as f:
                    f.write(caption_srt)
                logger.debug(f'completed writing caption file for {filepath}')
                is_file_exist = os.path.exists(filepath)
                logger.debug(f'file exist or not: {is_file_exist}')
            except Exception as e:
                logger.error(f'error for writing caption file: {str(e)}')
        except IndexError:
            pass



import os

from yt_concate_py.settings import DOWNLOADS_DIR
from yt_concate_py.settings import CAPTIONS_DIR
from yt_concate_py.settings import VIDEOS_DIR
from yt_concate_py.settings import OUTPUTS_DIR


class Utils:
    def __init__(self):
        pass

    def create_dirs(self):
        os.makedirs(DOWNLOADS_DIR, exist_ok=True)
        os.makedirs(CAPTIONS_DIR, exist_ok=True)
        os.makedirs(VIDEOS_DIR, exist_ok=True)
        os.makedirs(OUTPUTS_DIR, exist_ok=True)

    def get_video_list_filepath(self, channel_id):
        return os.path.join(DOWNLOADS_DIR, channel_id + '.txt')

    def video_list_file_exist(self, channel_id):
        path = self.get_video_list_filepath(channel_id)
        return os.path.exists(path) and os.path.getsize(path) > 0

    def get_output_video_filepath(self, channel_id, search_term):
        filename = channel_id + '_' + search_term + '.mp4'
        return os.path.join(OUTPUTS_DIR, filename)

    def get_temp_audio_filepath(self, channel_id, search_term):
        filename = channel_id + '_' + search_term + 'TEMP_MPY_wvf_snd.mp4'
        return os.path.join(OUTPUTS_DIR, filename)



from yt_concate_py.pipeline.steps.step import Step
from yt_concate_py.model.yt import YT
from yt_concate_py.logger import logger


class InitializeYT(Step):
    def process(self, data, inputs, utils):
        logger.info('INITIALIZING YT...')
        return [YT(video_link) for video_link in data]



from yt_concate_py.pipeline.steps.step import StepException


class Pipeline:
    def __init__(self, steps):
        self.steps = steps

    def run(self, inputs, utils):
        data = None
        for step in self.steps:
            try:
                data = step().process(data, inputs, utils)
            except StepException as e:
                print(e)
                break



import sys
import getopt
sys.path.append('../')
sys.path.append('/Users/cindy/Desktop/yt-concate/venv/lib/python3.9/site-packages')

from yt_concate_py.pipeline.pipeline import Pipeline
from yt_concate_py.pipeline.steps.prefligt import Preflight
from yt_concate_py.pipeline.steps.get_video_list import GetVideoList
from yt_concate_py.pipeline.steps.initialize_yt import InitializeYT
from yt_concate_py.pipeline.steps.download_captions import DownloadCaptions
from yt_concate_py.pipeline.steps.read_captions import ReadCaptions
from yt_concate_py.pipeline.steps.search import Search
from yt_concate_py.pipeline.steps.download_videos import DownloadVideos
from yt_concate_py.pipeline.steps.edit_video import EditVideo
from yt_concate_py.pipeline.steps.postflight import Postflight
from yt_concate_py.utils import Utils

CHANNEL_ID = 'UCKSVUHI9rbbkXhvAXK-2uxA'
SEARCH_TERM = 'incredible'
LIMIT = 10


def print_usage():
    print('python3 main.py -c <channel_id> -s <search_term> -l <limit>')
    print('python3 main.py --channel_id <channel_id> --search_term <search_term> --limit <limit>')

    print('python3 main.py OPTIONS')
    print('OPTIONS:')
    print('{:>10} {:<20}{}'.format('-c', '--channel_id', 'Channel id of YouTube channel to download videos'))
    print('{:>10} {:<20}{}'.format('-s', '--search_term', 'A search term used to extract video clips'))
    print('{:>10} {:<20}{}'.format('-l', '--limit', 'Maximum number of video clips to concatenate'))


def main():

    inputs = {
        'channel_id': CHANNEL_ID,
        'search_term': SEARCH_TERM,
        'limit': LIMIT,
    }

    short_opts = 'hc:s:l:'
    long_opts = 'help channel_id= search_term= limit='.split()

    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError:
        print_usage()
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print_usage()
            sys.exit(0)
        elif opt in ("-c", "--channel_id"):
            inputs['channel_id'] = arg
        elif opt in ("-s", "--search_term"):
            inputs['search_term'] = arg
        elif opt in ("-l", "--limit"):
            inputs['limit'] = arg
        else:
            print_usage()
            sys.exit(2)

    steps = [
                Preflight,
                GetVideoList,
                InitializeYT,
                DownloadCaptions,
                ReadCaptions,
                Search,
                DownloadVideos,
                EditVideo,
                Postflight,
            ]

    utils = Utils()

    p = Pipeline(steps)
    p.run(inputs, utils)


if __name__ == "__main__":
    main()



import time
import concurrent.futures

from pytubefix import YouTube
from pytubefix.cli import on_progress

from yt_concate_py.pipeline.steps.step import Step
from yt_concate_py.settings import VIDEOS_DIR
from yt_concate_py.logger import logger


class DownloadVideos(Step):
    def process(self, data, inputs, utils):
        logger.info('DOWNLOADING VIDEOS...')

        yt_set = set([found.yt for found in data])
        logger.info(f'prepare for downloading {len(yt_set)} videos...')

        yt_ = []
        for yt in yt_set:
            video_link = str(yt.video_link)

            if yt.video_file_exist:
                logger.debug(f'found existing video file : {video_link}, skipping...')
                continue
            else:
                yt_.append(yt)

        start_time = time.time()

        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
            executor.map(self.download_videos, yt_)

        end_time = time.time()
        elapsed_time = end_time - start_time
        logger.info('Completed downloading all videos')
        logger.info(f'Total running time for downloading videos : {elapsed_time} seconds')

        return data

    def download_videos(self, yt):
        video_link = str(yt.video_link)
        logger.debug(f'prepare to download yt video : {video_link}')
        youtube = YouTube(video_link, on_progress_callback=on_progress)
        ys = youtube.streams.get_highest_resolution()
        ys.download(output_path=VIDEOS_DIR, filename=yt.video_filename)
        logger.debug(f'completely downloaded yt video : {video_link}')



import urllib.request
import json

from yt_concate_py.pipeline.steps.step import Step
from yt_concate_py.settings import API_KEY
from yt_concate_py.logger import logger


class GetVideoList(Step):
    def process(self, data, inputs, utils):
        logger.info('GETTING VIDEO LIST...')
        channel_id = inputs['channel_id']
        api_key = API_KEY

        if utils.video_list_file_exist(channel_id):
            logger.debug('found existing video list file')
            return self.read_file(utils.get_video_list_filepath(channel_id))

        base_video_url = 'https://www.youtube.com/watch?v='
        base_search_url = 'https://www.googleapis.com/youtube/v3/search?'

        first_url = base_search_url + 'key={}&channelId={}&part=snippet,id&order=date&maxResults=25'.format(api_key,
                                                                                                            channel_id)

        video_links = []
        url = first_url

        while True:
            inp = urllib.request.urlopen(url)
            resp = json.load(inp)

            for i in resp['items']:
                if i['id']['kind'] == "youtube#video":
                    video_link = base_video_url + i['id']['videoId']
                    if video_link not in video_links:
                        video_links.append(video_link)

            try:
                next_page_token = resp['nextPageToken']
                url = first_url + '&pageToken={}'.format(next_page_token)
            except KeyError:
                break

        self.write_to_file(video_links, channel_id, utils)
        logger.info(f'{len(video_links)} videos from this channel')
        logger.debug(video_links)
        return video_links

    def write_to_file(self, video_links, channel_id, utils):
        filepath = utils.get_video_list_filepath(channel_id)
        with open(filepath, 'w') as f:
            for video_link in video_links:
                f.write(video_link + '\n')

    def read_file(self, filepath):
        video_links = []
        with open(filepath, 'r') as f:
            for line in f:
                video_links.append(line.strip())
        return video_links



from yt_concate_py.pipeline.steps.step import Step
from yt_concate_py.model.found import Found
from yt_concate_py.logger import logger


class Search(Step):
    def process(self, data, inputs, utils):
        logger.info('SEARCHING THE TERM...')
        search_term = inputs['search_term']
        found = []
        for yt in data:
            captions = yt.captions
            if not captions:
                continue
            for caption in captions:
                if search_term in caption:
                    time = captions[caption]
                    f = Found(yt, caption, time)
                    found.append(f)

        logger.info(f'found {len(found)} occurrences')
        logger.info(found)
        return found
